#include <windows.h>
#include <stdarg.h>
#include <time.h>
#include <stdio.h>
#include <tchar.h>
#include "service.h"
#include "loadsscvf.h"

#pragma comment(lib, "advapi32.lib")
#pragma comment(lib, "user32.lib")

#define MYNAME "Load & Clean SuperCache II"

BOOL g_bCheckOnly = FALSE;
BOOL g_bRemoveOnly = FALSE;
BOOL g_bLog = FALSE;
char g_szLogName[_MAX_FNAME] = "C:\\LoadSSCVF.log";
char g_szVolumeKey[260] = "SYSTEM\\CurrentControlSet\\Enum\\STORAGE\\Volume";
char g_szPrgName[80] = MYNAME;

void WriteLog(char *fmt, ...)
{
    char buf[1024];
    va_list marker;
    FILE *fp;
    time_t t;
    struct tm *tp;
    char tm_str[260];

    if (g_bLog == 0) return;
    va_start(marker, fmt);
    vsprintf(buf, fmt, marker);
    va_end(marker);

    time(&t);
    tp = localtime(&t);
    if (tp != NULL)
        sprintf(tm_str, "%04d/%02d/%02d %02d:%02d:%02d", tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday, tp->tm_hour, tp->tm_min, tp->tm_sec);
    else
        strcpy(tm_str, "[CAN'T GET LOCALTIME]");

    fp = fopen(g_szLogName, "at");
    if (fp != NULL) {
        fprintf(fp, "%s %s\n", tm_str, buf);
        fclose(fp);
    }
    return;
}

int ProcessEachVolume(char *szVolumeName, BOOL bRemove)
{
    HKEY hk;                      // registry key handle
    LONG nRC;
    DWORD dwIndex;
    char szKeyName[1024];
    char szName[1024];
    DWORD dwNameLen;
    char szMessage[1024];
    int nCount;

    WriteLog("Processing volume: %s", szVolumeName);
    sprintf(szKeyName, "%s\\%s\\Device Parameters\\SscVF", g_szVolumeKey, szVolumeName);
    nRC = RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKeyName, 0, KEY_ALL_ACCESS, &hk);
    if (nRC != ERROR_SUCCESS) {
        // we don't need to process it when the subkey not exist
        if (nRC == 2) {// not exist
            WriteLog("not exist, skip %s", szVolumeName);
            return 0;
        }
        sprintf(szMessage, "RegOpenKeyEx() failed %lx for HKLM %s", nRC, szKeyName);
        WriteLog(szMessage);
        AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
        MessageBox(NULL, szMessage, MYNAME, MB_OK);
        return 0;
    }

    nCount = 0;
    dwIndex = 0;
    while (1) {
        dwNameLen = 1024;
        nRC = RegEnumValue(hk, dwIndex, szName, &dwNameLen, NULL, NULL, NULL, NULL);
        if (nRC != ERROR_SUCCESS) {
            if (nRC == ERROR_NO_MORE_ITEMS)
                break;
            sprintf(szMessage, "RegEnumValue() failed %lx for %s, Value %s", nRC, szKeyName, szName);
            WriteLog(szMessage);
            AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
            MessageBox(NULL, szMessage, MYNAME, MB_OK);
            break;
        }
        nCount++;
        if (bRemove) {
            WriteLog("Delete Value: %s", szName);
            //dwIndex++;
            nRC = RegDeleteValue(hk, szName);
            if (nRC != ERROR_SUCCESS) {
                sprintf(szMessage, "RegDeleteValue() failed %lx for %s, Value %s", nRC, szKeyName, szName);
                WriteLog(szMessage);
                AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
                MessageBox(NULL, szMessage, MYNAME, MB_OK);
            }
        }
        else {
            WriteLog("Value exist: %s", szName);
            dwIndex++;
        }
    }
    RegCloseKey(hk);

    if (bRemove == FALSE && dwIndex > 0) {
        sprintf(szMessage, "SSCVF Value still exist %d, it may cause the system hang during booting!", dwIndex);
        WriteLog(szMessage);
        AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
        MessageBox(NULL, szMessage, MYNAME, MB_OK);
    }
    return nCount;
}

int CheckSSCVFKey(void)
{
    HKEY hk;
    LONG nRC;
    DWORD dwIndex;
    char szName[1024];
    DWORD dwNameLen;
    FILETIME ftTime;
    char szMessage[1024];

    nRC = RegOpenKeyEx(HKEY_LOCAL_MACHINE, g_szVolumeKey, 0, KEY_ENUMERATE_SUB_KEYS, &hk);
    if (nRC != ERROR_SUCCESS) {
        sprintf(szMessage, "RegOpenKeyEx() failed %lx for HKLM %s", nRC, g_szVolumeKey);
        WriteLog(szMessage);
        AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
        MessageBox(NULL, szMessage, MYNAME, MB_OK);
        return 1;
    }

    dwIndex = 0;
    while (1) {
        dwNameLen = 1024;
        nRC = RegEnumKeyEx(hk, dwIndex, szName, &dwNameLen, NULL, NULL, NULL, &ftTime);
        if (nRC != ERROR_SUCCESS) {
            if (nRC == ERROR_NO_MORE_ITEMS) break;
            sprintf(szMessage, "RegEnumKeyEx() failed %lx for HKLM %s", nRC, g_szVolumeKey);
            WriteLog(szMessage);
            AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
            MessageBox(NULL, szMessage, MYNAME, MB_OK);
            break;
        }
        ProcessEachVolume(szName, FALSE);
        dwIndex++;
    }
    RegCloseKey(hk);
    return 0;
}

int RemoveSSCVFKey(void)
{
    HKEY hk;
    LONG nRC;
    DWORD dwIndex;
    char szName[1024];
    DWORD dwNameLen;
    FILETIME ftTime;
    char szMessage[1024];
    int nCount;

    nRC = RegOpenKeyEx(HKEY_LOCAL_MACHINE, g_szVolumeKey, 0, KEY_ENUMERATE_SUB_KEYS, &hk);
    if (nRC != ERROR_SUCCESS) {
        sprintf(szMessage, "RegOpenKeyEx() failed %lx for HKLM %s", nRC, g_szVolumeKey);
        WriteLog(szMessage);
        AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
        MessageBox(NULL, szMessage, MYNAME, MB_OK);
        return 1;
    }

    nCount = 0;
    dwIndex = 0;
    while (1) {
        dwNameLen = 1024;
        nRC = RegEnumKeyEx(hk, dwIndex, szName, &dwNameLen, NULL, NULL, NULL, &ftTime);
        if (nRC != ERROR_SUCCESS) {
            if (nRC == ERROR_NO_MORE_ITEMS) break;
            sprintf(szMessage, "RegEnumKeyEx() failed %lx for HKLM %s", nRC, g_szVolumeKey);
            WriteLog(szMessage);
            AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
            MessageBox(NULL, szMessage, MYNAME, MB_OK);
            break;
        }
        if (ProcessEachVolume(szName, TRUE) > 0)
            nCount++;
        dwIndex++;
    }
    RegCloseKey(hk);
    return nCount;
}

//int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
int LoadSSCVF(void)
{
    char drive[_MAX_DRIVE];
    char dir[_MAX_DIR];
    char fname[_MAX_FNAME];
    char ext[_MAX_EXT];
    char szIniFile[_MAX_FNAME];
    int i, nMaxCommand;
    char szCommandKey[260];
    char szCommand[260];
    PROCESS_INFORMATION pi;
    STARTUPINFO si;
    char szMessage[1024];
    int nCount, nMaxTry, nSession;
    DWORD dwExitCode;

    _splitpath(__argv[0], drive, dir, fname, ext);
	sprintf(szIniFile, "%s%s%s.ini", drive, dir, fname);

    // do we need to generate log
    g_bLog = GetPrivateProfileInt(MYNAME, "WriteLog", 1, szIniFile);
    if (g_bLog) {
        // get log filename
        GetPrivateProfileString(MYNAME, "LogFile", "C:\\LoadSSCVF.log", g_szLogName, 260, szIniFile);
        if (GetPrivateProfileInt(MYNAME, "AppendLog", 0, szIniFile) == 0) {
            // if not append, delete it first
            DeleteFile(g_szLogName);
        }
    }

    // get volume key
    GetPrivateProfileString(MYNAME, "VolumeKey", "SYSTEM\\CurrentControlSet\\Enum\\STORAGE\\Volume", g_szVolumeKey, 260, szIniFile);
    WriteLog("VolumeKey = %s", g_szVolumeKey);

    nMaxTry = GetPrivateProfileInt(MYNAME, "MaxTry", 3, szIniFile);
    WriteLog("MaxTry = %d", nMaxTry);

    nSession = 0;
    nCount = 0;
    while (1) {
        nCount++;
        if (g_bRemoveOnly == FALSE && g_bCheckOnly == FALSE) {
            // get max command
            nMaxCommand = GetPrivateProfileInt(MYNAME, "MaxCommand", 1, szIniFile);
            WriteLog("MaxCommand = %d", nMaxCommand);
            for (i = 0; i < nMaxCommand; i++) {
                sprintf(szCommandKey, "SSCVF_Command_%d", i+1);
                // get command
                GetPrivateProfileString(MYNAME, szCommandKey, "", szCommand, 260, szIniFile);
                WriteLog("%s = %s", szCommandKey, szCommand);
                if (szCommand[0] == '\0') {
                    WriteLog("Empty command, skip for %s", szCommandKey);
                    continue;
                }

                ZeroMemory(&si, sizeof(si));
                si.cb = sizeof(si);
                si.wShowWindow = SW_HIDE;
                si.dwFlags = STARTF_USESHOWWINDOW;
                ZeroMemory(&pi, sizeof(pi));

                WriteLog("Execute: %s", szCommand);
                if (!CreateProcess(NULL, szCommand, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
                    sprintf(szMessage, "Execute failed: %s", szCommand);
                    WriteLog(szMessage);
                    AddToEventLog(EVENTLOG_ERROR_TYPE, szMessage);
                    MessageBox(NULL, szMessage, MYNAME, MB_OK);
                }
                else {
                    // Wait until child process exits.
                    WaitForSingleObject(pi.hProcess, INFINITE);

                    GetExitCodeProcess(pi.hProcess, &dwExitCode);
                    // Close process and thread handles.
                    CloseHandle(pi.hProcess);
                    CloseHandle(pi.hThread);
                    AddToEventLog(EVENTLOG_INFORMATION_TYPE, "Success (0x%lx) for %s", dwExitCode, szCommand);
                    WriteLog("Success (0x%lx) for %s", dwExitCode, szCommand);
                }
            }
        }

        nSession = 0;
        if (g_bCheckOnly == FALSE)
            nSession = RemoveSSCVFKey();

        if (g_bRemoveOnly == FALSE && g_bCheckOnly == FALSE) {
            if (nSession != nMaxCommand) {
                AddToEventLog(EVENTLOG_INFORMATION_TYPE, "nSession (%d) != nMaxCommand (%d), maybe not enable SuperCache II", nSession, nMaxCommand);
                WriteLog("nSession (%d) != nMaxCommand (%d), maybe not enable SuperCache II", nSession, nMaxCommand);
                if (nCount <= nMaxTry)
                    continue;
            }
        }
        break;
    }

    CheckSSCVFKey();

    if (GetPrivateProfileInt(MYNAME, "ShowFinished", 0, szIniFile) != 0)
        MessageBox(NULL, "Finished", MYNAME, MB_OK);

    return 0;
}

//
//  FUNCTION: ServiceStart
//
//  PURPOSE: Actual code of the service
//       that does the work.
//
//  PARAMETERS:
//    dwArgc   - number of command line arguments
//    lpszArgv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
VOID ServiceStart (DWORD dwArgc, LPTSTR *lpszArgv)
{
    BOOL bSuccess;
    char windir[260];
    char fname[260];

_TRY_EXCEPTION_BLOCK_BEGIN()

    if (g_bRemoveOnly || g_bCheckOnly) {
        LoadSSCVF();
        return;
    }

    ///////////////////////////////////////////////////
    //
    // Service initialization
    //

    bSuccess = FALSE;

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
                        SERVICE_START_PENDING, // service state
                        NO_ERROR,          // exit code
                        3000))             // wait hint
        goto cleanup;

    // report the status to the service control manager.
    //
    if (!ReportStatusToSCMgr(
                        SERVICE_RUNNING,       // service state
                        NO_ERROR,          // exit code
                        0))            // wait hint
        goto cleanup;

    bSuccess = TRUE;
    AddToEventLog(EVENTLOG_INFORMATION_TYPE, "%s Service Startup.\n", g_szPrgName);

    //
    // End of initialization
    //
    ////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////
    //
    // Service is now running, perform work until shutdown
    //

    LoadSSCVF();
    Sleep(3000);

cleanup:

    if (bSuccess)
        AddToEventLog(EVENTLOG_INFORMATION_TYPE, "%s Service Shutdown.\n", g_szPrgName);
    else
        AddToEventLog(EVENTLOG_ERROR_TYPE, "%s Service Startup Failed.\n", g_szPrgName);

_TRY_EXCEPTION_BLOCK_END(1)

    return;
}


//
//  FUNCTION: ServiceStop
//
//  PURPOSE: Stops the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    If a ServiceStop procedure is going to
//    take longer than 3 seconds to execute,
//    it should spawn a thread to execute the
//    stop code, and return.  Otherwise, the
//    ServiceControlManager will believe that
//    the service has stopped responding.
//
VOID ServiceStop()
{
_TRY_EXCEPTION_BLOCK_BEGIN()

_TRY_EXCEPTION_BLOCK_END(1)

    return;
}


VOID AfterServiceInstall()
{
_TRY_EXCEPTION_BLOCK_BEGIN()

_TRY_EXCEPTION_BLOCK_END(1)

    return;
}

VOID AfterServiceRemove()
{

_TRY_EXCEPTION_BLOCK_BEGIN()


_TRY_EXCEPTION_BLOCK_END(1)

    return;
}

